import os
from ament_index_python.packages import get_package_share_directory
from launch import LaunchDescription
from launch_ros.actions import Node
from launch.actions import DeclareLaunchArgument
from launch.substitutions import LaunchConfiguration, PathJoinSubstitution

def generate_launch_description():
    robot_package = os.environ['SCHPIN_PACKAGE']

    robot_urdf_filename = os.path.join(get_package_share_directory(robot_package), 'urdf/robot.urdf')
    with open(robot_urdf_filename, 'r') as fd:
        robot_urdf = fd.read()
    robot_srdf_filename = os.path.join(get_package_share_directory(robot_package), 'srdf/robot.srdf')
    with open(robot_srdf_filename, 'r') as fd:
        robot_srdf = fd.read()

    params = os.path.join(get_package_share_directory(robot_package), "config/control.yml")

    return LaunchDescription([
        Node(
            package='joint_state_publisher',
            executable='joint_state_publisher',
            output='both',
            parameters=[{
                "source_list" : ["/control/joint_states"]
            }]
        ),
        Node(package='robot_state_publisher',
             executable='robot_state_publisher',
            output='both',
             parameters=[{"robot_description": robot_urdf}]),
        Node(package='diagnostic_aggregator',
             executable='aggregator_node',
             output='both',
             parameters=[{
                "pub_rate": 1.,
                "base_path": '',
                "analyzers": {
                    "servomotors":{
                        "type":"diagnostic_aggregator/GenericAnalyzer",
                        "path": "Servomotors",
                        "startswith": ["Servomotor"]
                    }
                }
             }]),
        Node(
            package='schpin_control',
            executable='control_fw',
            name='control_fw',
            output='both',
            parameters=[
                {"robot_description": robot_urdf}
            ]
        ),
        DeclareLaunchArgument('script', default_value=None),
        Node(
            package='schpin_control',
            executable='control_node',
            name='control_node',
            output='both',
            #prefix=['sakura -e cgdb --args'],
            parameters=[params,{"robot_description": robot_urdf,
                "robot_semantics": robot_srdf,
                "script": LaunchConfiguration('script')
            }]
        )
    ])
