#include "fw/dynmx/protocol.h"

#pragma once

namespace schpin
{

struct dynmx_read_request
{
        std::size_t count;
};

template < typename Message >
class dynmx_sequencer
{
        // HEADER: 0xFF,0xFF,0xFD,0x00,<id>,<len_low>,<len_high>,<ins>

        enum states
        {
                IDLE,
                HEADER_START,
                HEADER_END,
                PACKET
        };

public:
        using message_type                                     = Message;
        static constexpr std::array< uint8_t, 4 > header_start = { 0xFF, 0xFF, 0xFD, 0x00 };
        static constexpr std::size_t              header_size  = 7;

private:
        states                                                               state_;
        std::array< uint8_t, em::max( Message::max_size, 2 * header_size ) > buffer_;
        std::size_t                                                          bi_ = 0;

public:
        dynmx_read_request start_read_packet()
        {
                state_ = HEADER_START;
                bi_    = 0;
                return { header_size };
        }

        template < typename Iterator >
        em::either< dynmx_read_request, Message > isr_read_callback( em::view< Iterator > data )
        {
                // TODO: nothing here checks bounds, debug that!
                std::size_t i     = 0;
                std::size_t end_i = bi_ + data.size();
                bool        found = false;
                std::copy( data.begin(), data.end(), buffer_.begin() + bi_ );
                switch ( state_ ) {
                        case IDLE:
                                return { dynmx_read_request{ 0 } };
                        case HEADER_START:
                                for ( ; i < header_size - header_start.size() + 1; ++i ) {
                                        auto to_check =
                                            em::view_n( buffer_.begin() + i, header_start.size() );
                                        if ( em::equal( to_check, header_start ) ) {
                                                found = true;
                                                break;
                                        }
                                }
                                if ( !found ) {
                                        std::copy( data.begin(), data.end(), buffer_.begin() );
                                        bi_ = header_size;
                                        return { dynmx_read_request{ header_size } };
                                }
                                std::copy(
                                    buffer_.begin() + i, buffer_.begin() + end_i, buffer_.begin() );
                                if ( i + header_size > end_i ) {
                                        state_ = HEADER_END;
                                        bi_    = end_i - i;
                                        return { dynmx_read_request{ header_size - bi_ } };
                                }
                                [[fallthrough]];
                        case HEADER_END:
                                state_ = PACKET;
                                bi_    = header_size;
                                return { dynmx_read_request{ get_size() } };
                        case PACKET:
                                state_ = IDLE;
                                bi_ += get_size();
                                return { Message{ buffer_, bi_ } };
                }
                return { dynmx_read_request{ 0 } };
        }

private:
        std::size_t get_size() const
        {
                // TODO:: this is not safe, use protocol<dynmx_packet_size>::deserialize
                return static_cast< std::size_t >( ( ( buffer_[6] << 8 ) + buffer_[5] ) );
        }
};
}  // namespace schpin
