#include "fw/lewan/protocol.h"

#pragma once

namespace schpin
{

struct lewan_read_request
{
        std::size_t count;
};

// TODO: dynmx_sequencer implements same mechanism: generalize!
template < typename Message >
class lewan_sequencer
{

        enum states
        {
                IDLE,
                HEADER_START,
                HEADER_END,
                PACKET
        };

public:
        using message_type                       = Message;
        static constexpr std::size_t header_size = 4;
        static constexpr std::size_t size_i      = 3;

private:
        states                                   state_;
        std::array< uint8_t, Message::max_size > buffer_;
        std::size_t                              bi_ = 0;

public:
        lewan_read_request start_read_packet()
        {
                state_ = HEADER_START;
                bi_    = 0;
                return { header_size };
        }

        template < typename Iterator >
        em::either< lewan_read_request, Message > isr_read_callback( em::view< Iterator > data )
        {
                // TODO: nothing here checks bounds, debug that!
                std::size_t i     = 0;
                bool        found = false;
                std::copy( data.begin(), data.end(), buffer_.begin() + bi_ );
                switch ( state_ ) {
                        case IDLE:
                                return { lewan_read_request{ 0 } };
                        case HEADER_START:
                                for ( ; i < header_size - 1; ++i ) {
                                        if ( buffer_[i] == *lewan_header_start_v &&
                                             buffer_[i + 1] == *lewan_header_start_v ) {
                                                found = true;
                                                break;
                                        }
                                }
                                if ( !found ) {
                                        if ( buffer_[header_size - 1] == *lewan_header_start_v ) {
                                                buffer_[0] = buffer_[header_size - 1];
                                                bi_        = 1;
                                        }
                                        return { lewan_read_request{ header_size - bi_ } };
                                }
                                if ( i != 0 ) {
                                        state_ = HEADER_END;
                                        bi_    = header_size - i;
                                        std::copy(
                                            buffer_.begin() + i,
                                            buffer_.begin() + header_size,
                                            buffer_.begin() );
                                        return { lewan_read_request{ header_size - bi_ } };
                                }
                                [[fallthrough]];
                        case HEADER_END:
                                state_ = PACKET;
                                bi_    = header_size;
                                return { lewan_read_request{
                                    static_cast< std::size_t >( buffer_[size_i] - 1 ) } };
                        case PACKET:
                                state_ = IDLE;
                                bi_ += static_cast< std::size_t >( buffer_[size_i] - 1 );
                                return Message{ buffer_, bi_ };
                }
                return { lewan_read_request{ 0 } };
        }
};
}  // namespace schpin
