#include "test/servo_runner.h"

using namespace schpin;

void servo_runner::visitor::to_servo( s_ctl_id msg_id, reply r, const master_lewan_message& m )
{
        SCHPIN_ASSERT( runner.ser_ );

        runner.ser_->write( m );
        if ( r == reply::NO ) {
                return;
        }
        runner.ser_->read_with_sequencer< lewan_sequencer< lewan_master_message > >().match(
            [&]( lewan_master_message m ) {
                    lewan_master_protocol::extract( m ).match(
                        em::uncurry |
                            [&]( lewan_header_start,
                                 lewan_header_start,
                                 lewan_servo_id,
                                 lewan_master_variant msg_var,
                                 lewan_checksum ) {
                                    auto clk_time = get_system_time();
                                    SCHPIN_INFO_LOG( "servo", "clktime: " << clk_time );
                                    runner.servo_->on_servo_receive( clk_time, msg_id, msg_var );
                            },
                        [&]( auto err ) {
                                SCHPIN_ERROR_LOG( "lewan_servo", "Lewan prot error: " << err );
                        } );
            },
            [&]( auto err ) {
                    SCHPIN_ERROR_LOG( "control", "Read error: " << err );
            } );
}

void servo_runner::visitor::to_servo( s_ctl_id msg_id, reply, const master_dynmx_message& m )
{
        SCHPIN_ASSERT( runner.ser_ );

        runner.ser_->write( m );

        runner.ser_->read_with_sequencer< dynmx_sequencer< dynmx_master_message > >()
            .convert_right( []( auto e ) {
                    return to_string( e );
            } )
            .bind_left( [&]( dynmx_master_message msg ) {
                    return dynmx_master_protocol::extract( msg ).convert_right( []( auto e ) {
                            return to_string( e );
                    } );
            } )
            .match(
                em::uncurry |
                    [&]( dynmx_header, dynmx_servo_id, auto tpl, dynmx_crc ) {
                            const dynmx_response_error& err = std::get< 1 >( tpl );
                            if ( err != dynmx_no_error ) {
                                    SCHPIN_ERROR_LOG( "runner", "Dynmx return error: " << err );
                            }
                            const dynmx_buffer& buffer = std::get< 2 >( tpl );

                            if ( buffer.size() != 0 ) {
                                    runner.servo_->on_servo_receive(
                                        get_system_time(),
                                        msg_id,
                                        servo_receive_variant{ buffer } );
                            }
                    },
                [&]( auto err ) {
                        SCHPIN_ERROR_LOG( "servo_runner", err << ", msg: " << m );
                } );
}

void servo_runner::tick()
{
        visitor vis{ *this };
        servo_->tick( get_system_time(), vis );
        servo_->poll( get_system_time(), 0, vis );
        std::string msg = repl_.prompt();
        vm_.load( msg ).optionally_log();
        vm_.full_exec( [&]( lua_State* lua_state, auto event ) {
                   return handle_event( lua_state, event );
           } )
            .optionally_log();
}
void servo_runner::run_file( const std::filesystem::path& file )
{
        auto opt_e = vm_.load( file );
        if ( opt_e ) {
                opt_e.optionally_log();
                return;
        }
        bool finished = false;

        visitor vis{ *this };
        while ( !finished ) {
                servo_->tick( get_system_time(), vis );
                servo_->poll( get_system_time(), 0, vis );
                vm_.exec( [&]( lua_State* lua_ptr, auto event ) {
                        return handle_event( lua_ptr, event );
                } )
                    << match(
                           [&]( lua_res res ) {
                                   finished = res == lua_res::FINISHED;
                           },
                           [&]( slua_error err ) {
                                   finished = true;
                                   SCHPIN_ERROR_LOG( "runner", "Slua error: " << err );
                           } );
        }
}
void servo_runner::handle_event( lua_State*, print_help_event )
{
        nlohmann::json res = slua_runner_def::get_def();
        print_help( std::cout, res );
}
void servo_runner::handle_event( lua_State*, print_config_event )
{
        servo_->dump_info();
}
nlohmann::json servo_runner::handle_event( lua_State* lua_ptr, msg_event e )
{
        nlohmann::json res;

        // TODO: write proper thing for this in schpin_lib/json and check in rest of the
        // codebase
        try {
                auto var = e.j.get< master_servo_command_variant >();

                servo_->on_servo_packet( ctl_id{ 0 }, var )
                    .match(
                        [&]( servo_events_variant var ) {
                                visitor vis{ *this };
                                servo_->on_event( ctl_id{ 0 }, get_system_time(), var, vis );
                        },
                        [&]( servo_master_message m ) {
                                servo_master_group::extract( m ).match(
                                    [&]( auto var ) {
                                            res = var;
                                    },
                                    [&]( auto err ) {
                                            SCHPIN_ERROR_LOG(
                                                "robot", "this should not happen error: " << err );
                                    } );
                        } );
        }
        catch ( const nlohmann::json::exception& ex ) {
                lua_error_stream( lua_ptr, "Json error: ", ex.what() );
        }
        return res;
}
