#include "fw/dynmx/sequencer.h"
#include "fw/lewan/sequencer.h"
#include "fw/serial.h"
#include "fw/servo/interface.h"
#include "fw/util.h"
#include "node/lua_vm/extra_ser.h"
#include "test/servo_runner/lua_def.h"

#include <schpin_lib/bucket_memory_resource.h>
#include <schpin_lib/json/tuple_variant_serializer.h>
#include <schpin_lib/lua/help.h>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/util.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

#pragma once

template <>
struct nlohmann::adl_serializer< schpin::master_servo_command_variant >
  : schpin::tuple_variant_serializer< schpin::master_servo_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< schpin::servo_master_command_variant >
  : schpin::tuple_variant_serializer< schpin::servo_master_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< schpin::servo_error_variant >
  : schpin::tuple_variant_serializer< schpin::servo_error_variant >
{
};

namespace schpin
{

template <>
struct ros_param_serializer< servo_type > : ros_param_serializer_enum< servo_type >
{
};

using runner_error = error< struct runner_error_tag >;

struct servo_tester_config
{
        servo_type                             type;
        std::optional< servo_id >              opt_id;
        std::optional< std::filesystem::path > opt_dev;
        std::optional< std::filesystem::path > opt_source;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_param< servo_type >( "stype", "Servo type" ),
                    ros_param< std::optional< servo_id > >( "sid", "Servo id" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "device", "Device to open for connection" ),
                    ros_param< std::optional< std::filesystem::path > >(
                        "script", "Script to execute instead of the interactive mode" ) );
        }
};

class servo_runner
{
        using mem = bucket_memory_resource< 1, servo_bucket_size >;

        mem                                memory_;
        std::optional< serial_connection > ser_;
        servo_ptr                          servo_;
        slua_vm< slua_runner_def >         vm_;
        slua_repl                          repl_;

        struct visitor : servo_visitor
        {
                servo_runner& runner;

                visitor( servo_runner& srunner )
                  : runner( srunner )
                {
                }

                void to_control( servo_id, const servo_master_message& )
                {
                }

                void on_activate( servo_id id )
                {
                        SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                }

                void to_servo( s_ctl_id msg_id, reply r, const master_lewan_message& m );

                void to_servo( s_ctl_id msg_id, reply, const master_dynmx_message& m );
        };

public:
        servo_runner(
            const servo_tester_config&         conf,
            std::optional< serial_connection > opt_ser = {} )
          : ser_( opt_ser )
          , servo_( make_servo( conf.type, *conf.opt_id, get_system_time(), &memory_ ) )
        {
                SCHPIN_ASSERT( conf.opt_id );
                visitor vis{ *this };
                while ( servo_->is_activating() ) {
                        servo_->tick( get_system_time(), vis );
                }
        }

        void tick();

        void run_file( const std::filesystem::path& file );

        void handle_event( lua_State*, print_help_event );

        void handle_event( lua_State*, quit_event e )
        {
                std::exit( e.res );
        }

        void handle_event( lua_State*, print_config_event );

        nlohmann::json handle_event( lua_State* lua_ptr, msg_event e );
};

}  // namespace schpin
