#include "stm32g4xx_hal.h"

#include <cstring>
#include <cstdlib>

#pragma once

namespace schpin
{

struct perif_handlers
{
        FDCAN_HandleTypeDef can;
        I2C_HandleTypeDef   i2c;
        UART_HandleTypeDef  bus_uart;
        UART_HandleTypeDef  com_uart;
};

inline void setup_oscillators()
{
        RCC_OscInitTypeDef osc_struct;
        memset(&osc_struct, 0, sizeof(RCC_OscInitTypeDef));

        osc_struct.OscillatorType      = RCC_OSCILLATORTYPE_HSI;
        osc_struct.HSIState            = RCC_HSI_ON;
        osc_struct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
        osc_struct.PLL.PLLState        = RCC_PLL_ON;
        osc_struct.PLL.PLLSource       = RCC_PLLSOURCE_HSI;
        osc_struct.PLL.PLLM            = RCC_PLLM_DIV4;
        osc_struct.PLL.PLLN            = 85;
        osc_struct.PLL.PLLP            = RCC_PLLP_DIV2;
        osc_struct.PLL.PLLQ            = RCC_PLLQ_DIV2;
        osc_struct.PLL.PLLR            = RCC_PLLR_DIV2;

        if ( HAL_RCC_OscConfig( &osc_struct ) != HAL_OK ) {
                std::abort();
        }
}

inline void setup_clock()
{
        RCC_ClkInitTypeDef clk_struct;
        memset(&clk_struct, 0, sizeof(RCC_ClkInitTypeDef));

        clk_struct.ClockType =
            RCC_CLOCKTYPE_HCLK | RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2;
        clk_struct.SYSCLKSource   = RCC_SYSCLKSOURCE_PLLCLK;
        clk_struct.AHBCLKDivider  = RCC_SYSCLK_DIV1;
        clk_struct.APB1CLKDivider = RCC_HCLK_DIV1;
        clk_struct.APB2CLKDivider = RCC_HCLK_DIV2;

        if ( HAL_RCC_ClockConfig( &clk_struct, FLASH_LATENCY_4 ) != HAL_OK ) {
                std::abort();
        }
}

inline void setup_periph_clock()
{
        RCC_PeriphCLKInitTypeDef clk_struct;
        memset(&clk_struct, 0, sizeof(RCC_PeriphCLKInitTypeDef));

        clk_struct.PeriphClockSelection =
            RCC_PERIPHCLK_USART1 | RCC_PERIPHCLK_USART2 | RCC_PERIPHCLK_I2C1 | RCC_PERIPHCLK_FDCAN;
        clk_struct.Usart1ClockSelection = RCC_USART1CLKSOURCE_PCLK2;
        clk_struct.Usart2ClockSelection = RCC_USART2CLKSOURCE_PCLK1;
        clk_struct.I2c1ClockSelection   = RCC_I2C1CLKSOURCE_PCLK1;

        if ( HAL_RCCEx_PeriphCLKConfig( &clk_struct ) != HAL_OK ) {
                std::abort();
        }
}

inline void setup_gpio()
{
        __HAL_RCC_GPIOA_CLK_ENABLE();
        __HAL_RCC_GPIOB_CLK_ENABLE();

        HAL_GPIO_WritePin( GPIOB, GPIO_PIN_8, GPIO_PIN_RESET );

        GPIO_InitTypeDef gpio_struct;

        gpio_struct.Pin   = GPIO_PIN_8;
        gpio_struct.Mode  = GPIO_MODE_OUTPUT_PP;
        gpio_struct.Pull  = GPIO_NOPULL;
        gpio_struct.Speed = GPIO_SPEED_FREQ_LOW;

        HAL_GPIO_Init( GPIOB, &gpio_struct );
}

inline FDCAN_HandleTypeDef setup_fdcan()
{
        FDCAN_HandleTypeDef res;

        res.Instance                  = FDCAN1;
        res.Init.ClockDivider         = FDCAN_CLOCK_DIV1;
        res.Init.FrameFormat          = FDCAN_FRAME_CLASSIC;
        res.Init.Mode                 = FDCAN_MODE_NORMAL;
        res.Init.AutoRetransmission   = DISABLE;
        res.Init.TransmitPause        = DISABLE;
        res.Init.ProtocolException    = DISABLE;
        res.Init.NominalPrescaler     = 1;
        res.Init.NominalSyncJumpWidth = 1;
        res.Init.NominalTimeSeg1      = 2;
        res.Init.NominalTimeSeg2      = 2;
        res.Init.DataPrescaler        = 1;
        res.Init.DataSyncJumpWidth    = 1;
        res.Init.DataTimeSeg1         = 1;
        res.Init.DataTimeSeg2         = 1;
        res.Init.StdFiltersNbr        = 0;
        res.Init.ExtFiltersNbr        = 0;
        res.Init.TxFifoQueueMode      = FDCAN_TX_FIFO_OPERATION;

        if ( HAL_FDCAN_Init( &res ) != HAL_OK ) {
                std::abort();
        }

        return res;
}

inline I2C_HandleTypeDef setup_i2c()
{
        I2C_HandleTypeDef res;

        res.Instance              = I2C1;
        res.Init.Timing           = 0x30A0A7FB;
        res.Init.OwnAddress1      = 0;
        res.Init.AddressingMode   = I2C_ADDRESSINGMODE_7BIT;
        res.Init.DualAddressMode  = I2C_DUALADDRESS_DISABLE;
        res.Init.OwnAddress2      = 0;
        res.Init.OwnAddress2Masks = I2C_OA2_NOMASK;
        res.Init.GeneralCallMode  = I2C_GENERALCALL_DISABLE;
        res.Init.NoStretchMode    = I2C_NOSTRETCH_DISABLE;

        if ( HAL_I2C_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_I2CEx_ConfigAnalogFilter( &res, I2C_ANALOGFILTER_ENABLE ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_I2CEx_ConfigDigitalFilter( &res, 0 ) != HAL_OK ) {
                std::abort();
        }

        return res;
}

inline UART_HandleTypeDef setup_lewan_uart()
{
        UART_HandleTypeDef res;
        res.Instance            = USART1;
        res.Init.BaudRate       = 115200;
        res.Init.WordLength     = UART_WORDLENGTH_8B;
        res.Init.StopBits       = UART_STOPBITS_1;
        res.Init.Parity         = UART_PARITY_NONE;
        res.Init.Mode           = UART_MODE_TX_RX;
        res.Init.HwFlowCtl      = UART_HWCONTROL_NONE;
        res.Init.OverSampling   = UART_OVERSAMPLING_16;
        res.Init.OneBitSampling = UART_ONE_BIT_SAMPLE_DISABLE;
        res.Init.ClockPrescaler = UART_PRESCALER_DIV1;
        res.AdvancedInit.AdvFeatureInit =
            UART_ADVFEATURE_TXINVERT_INIT | UART_ADVFEATURE_RXINVERT_INIT;
        res.AdvancedInit.TxPinLevelInvert = UART_ADVFEATURE_TXINV_ENABLE;
        res.AdvancedInit.RxPinLevelInvert = UART_ADVFEATURE_RXINV_ENABLE;

        if ( HAL_HalfDuplex_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetTxFifoThreshold( &res, UART_TXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetRxFifoThreshold( &res, UART_RXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_DisableFifoMode( &res ) != HAL_OK ) {
                std::abort();
        }
        return res;
}

inline UART_HandleTypeDef setup_com_uart()
{
        UART_HandleTypeDef res;
        res.Instance                    = USART2;
        res.Init.BaudRate               = 115200;
        res.Init.WordLength             = UART_WORDLENGTH_8B;
        res.Init.StopBits               = UART_STOPBITS_1;
        res.Init.Parity                 = UART_PARITY_NONE;
        res.Init.Mode                   = UART_MODE_TX_RX;
        res.Init.HwFlowCtl              = UART_HWCONTROL_NONE;
        res.Init.OverSampling           = UART_OVERSAMPLING_16;
        res.Init.OneBitSampling         = UART_ONE_BIT_SAMPLE_DISABLE;
        res.Init.ClockPrescaler         = UART_PRESCALER_DIV1;
        res.AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT;

        if ( HAL_UART_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_HalfDuplex_Init( &res ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetTxFifoThreshold( &res, UART_TXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_SetRxFifoThreshold( &res, UART_RXFIFO_THRESHOLD_1_8 ) != HAL_OK ) {
                std::abort();
        }
        if ( HAL_UARTEx_DisableFifoMode( &res ) != HAL_OK ) {
                std::abort();
        }
        return res;
}

inline perif_handlers setup_system()
{
        HAL_Init();

        setup_oscillators();
        setup_clock();
        setup_periph_clock();

        setup_gpio();
        return {
            .can      = setup_fdcan(),
            .i2c      = setup_i2c(),
            .bus_uart = setup_lewan_uart(),
            .com_uart = setup_com_uart() };
}

}  // namespace schpin
