#include "schpin_lib_exp/convert_iterator.h"
#include "schpin_lib_exp/stringify.h"

#pragma once

namespace schpin
{

template < typename T >
struct protocol_stringify
{
        static std::string make( const T& item )
        {
                std::stringstream ss;
                ss << item;
                return ss.str();
        }
};

template < typename MsgCounterType, std::size_t N, typename Owner >
struct protocol_stringify< protocol_recursive_promoter< prot_message< MsgCounterType, N, Owner > > >
{

        static std::string make( const prot_message< MsgCounterType, N, Owner >& item )
        {
                return Owner::stringify( item );
        }
};

template < typename T, T Min, T Max >
struct protocol_stringify< bounded< T, Min, Max > >
{
        static std::string make( const bounded< T, Min, Max >& item )
        {
                std::stringstream ss;
                ss << std::to_string( *item );
                return ss.str();
        }
};

template < typename T, std::size_t N >
struct protocol_stringify< std::array< T, N > >
{
        static std::string make( const std::array< T, N >& data )
        {
                std::stringstream ss;
                for ( const T& item : data ) {
                        ss << item << "  ";
                }
                return ss.str();
        }
};


template < typename MsgCounterType = uint8_t, typename... Commands >
struct protocol_stringify< command_group< MsgCounterType, Commands... > >
{
        using item            = command_group< MsgCounterType, Commands... >;
        using command_variant = typename item::command_variant;
        using message_t       = typename item::message_t;

        static std::string make( const command_variant& v )
        {
                using namespace stringify;

                std::stringstream ss;
                ss << v;
                return ss.str();
        }

        static std::string make( const message_t& m )
        {
                std::stringstream ss;

                item::extract( m ).match(
                    [&]( command_variant v ) {
                            ss << " [ " << make( v ) << " :: " << convert_view< int >( m ) << " ] ";
                    },
                    [&]( auto ) {
                            ss << em::view{ m };
                    } );
                return ss.str();
        }
};

template < typename MsgCounterType = uint8_t, typename... Items >
struct protocol_stringify< protocol_tuple< MsgCounterType, Items... > >
{
        using item       = protocol_tuple< MsgCounterType, Items... >;
        using input_type = typename item::input_type;
        using message_t  = typename item::message_t;

        static std::string stringify( const input_type& v )
        {
                std::stringstream ss;
                em::for_each(
                    em::make_sequence_tuple_t< sizeof...( Items ) >{}, [&]( auto integ_const ) {
                            constexpr std::size_t i = decltype( integ_const )::value;
                            using t                 = std::variant_alternative_t< i, item_tuple >;

                            ss << protocol_stringify< t >::make( std::get< i >( v ) ) << " | ";
                    } );

                return ss.str();
        }

        static std::string stringify( const message_t& m )
        {
                std::stringstream ss;
                item::extract( m ).match(
                    [&]( const input_type& v ) {
                            ss << "{ " << make( v ) << " || " << convert_view< int >( m ) << " }";
                    },
                    [&]( auto ) {
                            ss << convert_view< int >( m );
                    } );
                return ss.str();
        }
};

}  // namespace schpin
