// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "schpin_lib_exp/overload.h"
#include "schpin_lib_exp/protocol/item.h"
#include "schpin_lib_exp/static_view.h"

#pragma once

namespace schpin
{

template < typename MsgCounterType = uint8_t, typename... Items >
struct protocol_tuple
{

        template < typename NewMsgCounterType >
        using with_msg_counter = protocol_tuple< NewMsgCounterType, Items... >;

        template < typename... NewItems >
        using with_items = protocol_tuple< MsgCounterType, Items..., NewItems... >;

        template < typename... NewItems >
        using with_promoted_items =
            protocol_tuple< MsgCounterType, Items..., protocol_recursive_promoter< NewItems >... >;

        using item_tuple  = std::tuple< Items... >;
        using input_type  = typename protocol_item< item_tuple >::input_type;
        using output_type = typename protocol_item< item_tuple >::output_type;

        static constexpr uint16_t    item_n = static_cast< uint16_t >( sizeof...( Items ) );
        static constexpr std::size_t size   = ( protocol_item< Items >::size + ... + 0 );
        using message_t                     = prot_message< MsgCounterType, size, protocol_tuple >;

        template < typename... Args >
        static constexpr output_type make_val( Args&&... args )
        {
                return output_type{ args... };
        }

        template < typename... Args >
        static constexpr message_t create( Args&&... items_val )
        {
                return serialize( items_val... );
        }

        template < typename... Args >
        static constexpr message_t serialize( Args&&... items_val )
        {
                std::array< uint8_t, size > buff = { 0 };
                std::size_t used = serialize_at( make_static_view( buff ), items_val... );
                return message_t{ buff, used };
        }

        template < typename... Args >
        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > target, Args&&... items_val )
        {
                return protocol_item< item_tuple >::serialize_at(
                    target, input_type{ items_val... } );
        }

        static em::either< output_type, protocol_error_record >
        extract( const em::view< const uint8_t* >& m )
        {
                prot_res< output_type > res = protocol_item< item_tuple >::deserialize( m );
                return std::visit(
                    []( auto item ) {
                            return em::either< output_type, protocol_error_record >{ item };
                    },
                    res.val_var );
        }
};

}  // namespace schpin
